package main

import (
	"context"
	"net"

	"gitlab.mashplace.com/mashplace/backend/common/api/health"
	"gitlab.mashplace.com/mashplace/backend/common/generic/addr"
	"gitlab.mashplace.com/mashplace/backend/common/generic/env"
	"gitlab.mashplace.com/mashplace/backend/common/generic/grpc"
	logger "gitlab.mashplace.com/mashplace/backend/common/generic/log"
	"gitlab.mashplace.com/mashplace/backend/common/generic/metrics"
	"gitlab.mashplace.com/mashplace/backend/common/generic/tracer"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/extras/internal/model/mongo"
	"gitlab.mashplace.com/mashplace/backend/extras/internal/service"
	"gitlab.mashplace.com/mashplace/backend/extras/version"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
)

var (
	envMongoURL      = env.ResolveString("MONGO_URL", "mongodb:27017")
	envMongoUsername = env.ResolveString("MONGO_USERNAME", "extras-data")
	envMongoPassword = env.ResolveString("MONGO_PASSWORD", "extras-data")
	envMongoAuthDB   = env.ResolveString("MONGO_AUTH", "admin")
	envMongoDB       = env.ResolveString("MONGO_DATABASE", "extras-data")
	envMongoSSLMode  = env.ResolveBool("MONGO_SSL_MODE", false)
)

var (
	connectPropertyService = property.NewPropertyServiceClient(grpc.Dial(
		env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000"),
		"Property",
	))

	connectOfferService = offer.NewOfferServiceClient(grpc.Dial(
		env.ResolveString("OFFER_SERVICE_URL", "mashroom-offer:10000"),
		"Offer",
	))
)

func main() {
	ctx := context.Background()

	defer tracer.Close()

	repository, err := mongo.NewRepository(ctx, &mongo.Config{
		URL:          envMongoURL,
		Username:     envMongoUsername,
		Password:     envMongoPassword,
		Database:     envMongoDB,
		AuthSource:   envMongoAuthDB,
		MongoSSLMode: envMongoSSLMode,
	})
	if err != nil {
		panic(err)
	}

	err = repository.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	srv, err := service.NewService(ctx, &service.Config{
		Repository: repository,
		Connections: service.Connections{
			Property: connectPropertyService,
			Offer:    connectOfferService,
		},
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("extras")
	if err != nil {
		panic(err)
	}

	extras.RegisterExtrasServiceServer(grpcServer, srv)

	logger.Infoln("extras started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
