package service

import (
	"context"
	"fmt"
	"math"

	"gitlab.mashplace.com/mashplace/backend/common/api/grpcctx"
	"gitlab.mashplace.com/mashplace/backend/common/api/grpcerr"
	extras "gitlab.mashplace.com/mashplace/backend/extras/api"
	"gitlab.mashplace.com/mashplace/backend/extras/internal/model"
	offer "gitlab.mashplace.com/mashplace/backend/offer/api"
	property "gitlab.mashplace.com/mashplace/backend/property/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"google.golang.org/protobuf/types/known/wrapperspb"
)

var (
	nilError            = grpcerr.New(grpcerr.InvalidData, "nil")
	codeUniqueError     = grpcerr.New(grpcerr.InvalidData, "code already exist")
	codePredefinedError = grpcerr.New(grpcerr.InvalidData, "predefined codes cannot be changed or removed")
	codeNilError        = grpcerr.New(grpcerr.InvalidData, "code empty")
)

type serviceImpl struct {
	config Config
}

func (srv *serviceImpl) Add(ctx context.Context, request *extras.AddRequest) (*extras.AddResponse, error) {
	if request == nil {
		return nil, nilError
	}

	if request.GetExtraService() == nil {
		return nil, nilError
	}

	if request.GetExtraService() == nil {
		return nil, codeNilError
	}

	m := &model.Service{
		ID: primitive.NewObjectID(),
	}

	err := srv.updateService(ctx, m, request.GetExtraService())
	if err != nil {
		return nil, err
	}

	err = srv.config.Repository.ServiceCreate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &extras.AddResponse{
		Id: m.ID.Hex(),
	}, nil
}

func (srv *serviceImpl) Update(ctx context.Context, request *extras.UpdateRequest) (*extras.UpdateResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetExtraService().GetIdValue())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.ServiceGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	err = srv.updateService(ctx, m, request.GetExtraService())
	if err != nil {
		return nil, err
	}

	err = srv.config.Repository.ServiceUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &extras.UpdateResponse{Success: true}, nil
}

func (srv *serviceImpl) Get(ctx context.Context, request *extras.GetRequest) (*extras.GetResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.ServiceGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	prop, err := srv.resolveProperty(ctx, request.GetPropertyId(), request.GetProperty())
	if err != nil {
		return nil, err
	}

	svc, err := srv.renderPropertyService(prop, m, true, false)
	if err != nil {
		return nil, err
	}

	return &extras.GetResponse{ExtraService: svc}, nil
}

func (srv *serviceImpl) List(ctx context.Context, request *extras.ListRequest) (*extras.ListResponse, error) {
	if request.GetRole() == extras.ServiceRole_LANDLORD_LISTING {
		request.Role = extras.ServiceRole_LANDLORD_SERVICES
	}

	resp, err := srv.config.Repository.ServicesList(ctx, &model.ServicesListOptions{
		Role:        request.GetRole(),
		Type:        request.GetType(),
		CartProduct: request.GetCartProduct(),
		Attribute:   request.GetAttribute(),
	})
	if err != nil {
		return nil, err
	}

	prop, err := srv.resolveProperty(ctx, request.GetPropertyId(), request.GetProperty())
	if err != nil {
		return nil, err
	}

	results := srv.renderPropertyServices(prop, resp, true, true)

	return &extras.ListResponse{ExtraServices: results}, nil
}

func (srv *serviceImpl) ListWithPrice(ctx context.Context, request *extras.ListRequest) (*extras.ListResponse, error) {
	return srv.List(ctx, request)
}

func (srv *serviceImpl) Delete(ctx context.Context, request *extras.DeleteRequest) (*extras.DeleteResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.ServiceGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if srv.predefinedCode(m.Code) {
		return nil, codePredefinedError
	}

	err = srv.config.Repository.ServiceRemove(ctx, id)
	if err != nil {
		return nil, err
	}

	return &extras.DeleteResponse{Success: true}, nil
}

func (srv *serviceImpl) Search(ctx context.Context, request *extras.SearchRequest) (*extras.SearchResponse, error) {
	for _, p := range request.GetParams().GetFilter() {
		vs := p.GetValue()

		for i, v := range p.GetValue() {
			if v == extras.ServiceRole_LANDLORD_LISTING.String() {
				vs[i] = extras.ServiceRole_LANDLORD_SERVICES.String()
			}
		}

		p.Value = vs
	}

	svcs, total, err := srv.config.Repository.ServicesFilter(ctx, request.GetParams())
	if err != nil {
		return nil, err
	}

	prop, err := srv.resolveProperty(ctx, "", request.GetProperty())
	if err != nil {
		return nil, err
	}

	results := srv.renderPropertyServices(prop, svcs, false, true)

	return &extras.SearchResponse{
		Services: results,
		Total:    int32(total),
	}, nil
}

func (srv *serviceImpl) GetByCode(
	ctx context.Context,
	request *extras.GetByCodeRequest,
) (*extras.GetByCodeResponse, error) {
	if request.GetCode() == "" {
		return nil, fmt.Errorf("%w: service code is empty", ErrInvalidArgument)
	}

	s, err := srv.config.Repository.ServiceGetByCode(ctx, request.GetCode())
	if err != nil {
		return nil, err
	}

	return &extras.GetByCodeResponse{
		Service: srv.renderService(s),
	}, nil
}

func (srv *serviceImpl) PriceCalculate(
	ctx context.Context,
	request *extras.PriceCalculateRequest,
) (*extras.PriceCalculateResponse, error) {
	serviceID, err := primitive.ObjectIDFromHex(request.GetServiceId())
	if err != nil {
		return nil, err
	}

	service, err := srv.config.Repository.ServiceGetByID(ctx, serviceID)
	if err != nil {
		return nil, err
	}

	err = srv.serviceCalculate(request.GetProperty(), service)
	if err != nil {
		return nil, err
	}

	return &extras.PriceCalculateResponse{
		Price:     service.Price,
		TaxAmount: service.TaxAmount,
	}, nil
}

func (srv *serviceImpl) predefinedCode(s string) bool {
	for _, k := range extras.KnownProductCodes {
		if s == k.String() {
			return true
		}
	}

	return false
}

func (srv *serviceImpl) checkServiceCode(ctx context.Context, m *model.Service, s *extras.Service) error {
	if s.GetCodeValue() == "" {
		return codeNilError
	}

	if s.GetCodeValue() != m.Code {
		exists, err := srv.config.Repository.ServiceExistsByCode(ctx, s.GetCodeValue())
		if err != nil {
			return err
		}

		if exists {
			return codeUniqueError
		}
	}

	if srv.predefinedCode(m.Code) {
		return codePredefinedError
	}

	return nil
}

func parseServiceRecievers(svcs []*extras.ServiceReceiver) (serviceReceivers []*model.ServiceReceiver) {
	for _, r := range svcs {
		serviceReceivers = append(serviceReceivers, &model.ServiceReceiver{
			Code:    r.Code,
			Percent: r.Percent,
			Sum:     r.Sum,
		})
	}

	return
}

func (srv *serviceImpl) updateService(ctx context.Context, m *model.Service, s *extras.Service) error {
	if s.GetCode() != nil {
		err := srv.checkServiceCode(ctx, m, s)
		if err != nil {
			return err
		}

		m.Code = s.GetCodeValue()
	}

	if s.GetType() != nil {
		m.Type = s.GetTypeValue()
	}

	if s.GetRole() != nil {
		m.Role = s.GetRoleValue()
	}

	if s.GetStatus() != nil {
		m.Status = s.GetStatusValue()
	}

	if s.GetDescription() != nil {
		m.Description = s.GetDescriptionValue()
	}

	if s.GetDisabled() != nil {
		m.Disabled = s.GetDisabledValue()
	}

	if s.GetPrimary() != nil {
		m.Primary = s.GetPrimaryValue()
	}

	if s.GetTitle() != nil {
		m.Title = s.GetTitleValue()
	}

	if s.GetPriceNeedCalculate() != nil {
		m.PriceNeedCalculate = s.GetPriceNeedCalculateValue()
	}

	if s.GetPriceCalculated() != nil {
		m.PriceCalculated = s.GetPriceCalculatedValue()
	}

	if s.GetObligatory() != nil {
		m.Obligatory = s.GetObligatoryValue()
	}

	if s.GetPriceOnRequest() != nil {
		m.PriceOnRequest = s.GetPriceOnRequestValue()
	}

	if s.GetReferenceProvider() != nil {
		m.ReferenceProvider = s.GetReferenceProviderValue()
	}

	if s.GetReferenceId() != nil {
		m.ReferenceID = s.GetReferenceIdValue()
	}

	if s.GetServiceReceivers() != nil {
		m.ServiceReceivers = parseServiceRecievers(s.GetServiceReceiversValue().GetServiceReceivers())
	}

	if s.GetCartProduct() != nil {
		m.CartProduct = s.GetCartProductValue()
	}

	if s.GetPageUrl() != nil {
		m.PageURL = s.GetPageUrlValue()
	}

	if s.GetMainIconUrl() != nil {
		m.MainIconURL = s.GetMainIconUrlValue()
	}

	if s.GetInventoryCode() != nil {
		m.InventoryCode = s.GetInventoryCodeValue()
	}

	if s.GetInventoryAccount() != nil {
		m.InventoryAccount = s.GetInventoryAccountValue()
	}

	if s.GetInventoryOffset() != nil {
		m.InventoryOffset = s.GetInventoryOffsetValue()
	}

	if s.GetAttributesValue() != nil {
		m.Attributes = s.GetAttributesValue().AsMap()
	}

	srv.updateTax(m, s)

	return nil
}

func (srv *serviceImpl) updateTax(m *model.Service, s *extras.Service) {
	if s.GetTaxType() != nil {
		m.TaxType = s.GetTaxTypeValue()
	}

	if s.GetPrice() != nil {
		m.Price = s.GetPriceValue()
		m.TaxAmount = math.Round(m.TaxPercentage*m.Price) / 100
	}

	if s.GetTaxAmount() != nil {
		m.TaxAmount = s.GetTaxAmountValue()
	}

	if s.GetTaxPercentage() != nil {
		m.TaxPercentage = s.GetTaxPercentageValue()
		if m.TaxAmount == 0 || (s.GetTaxAmountValue() == 0 && s.GetTaxPercentageValue() != m.TaxPercentage) {
			m.TaxAmount = math.Round(m.Price*m.TaxPercentage) / 100
		}
	}
}

//nolint:staticcheck
func (srv *serviceImpl) resolveProperty(
	ctx context.Context,
	id string,
	prop *extras.Property,
) (*extras.Property, error) {
	if prop == nil {
		prop = &extras.Property{}
	}

	if prop.GetId() == "" {
		prop.Id = id
	}

	if prop.Details == nil {
		prop.Details = &property.PropertyDetails{}
	}

	if prop.GetId() != "" {
		resp, err := srv.config.Connections.Property.Get(grpcctx.WithServiceUser(ctx), &property.GetRequest{
			Id: prop.GetId(),
		})
		if err != nil {
			return nil, err
		}

		if prop.GetConditions() == nil || prop.GetConditions().GetPrice() == 0 {
			off := srv.rangeOffers(ctx, prop)
			if off != nil {
				prop.Conditions = &extras.Conditions{
					Price: off.GetConditions().GetPrice().GetValue(),
				}
			}
		}

		nprop := &extras.Property{
			Id:         resp.GetProperty().GetId(),
			Status:     resp.GetProperty().GetStatus(),
			Details:    resp.GetProperty().GetDetails(),
			Rental:     resp.GetProperty().GetRental(),
			People:     resp.GetProperty().GetPeople(),
			Conditions: prop.GetConditions(),
		}

		if nprop.Details == nil {
			nprop.Details = prop.GetDetails()
		}

		if nprop.GetDetails().BedroomsNo == nil {
			nprop.GetDetails().BedroomsNo = prop.GetDetails().GetBedroomsNo()
		}

		if nprop.GetDetails().GetAmenityFurniture() == 0 {
			nprop.GetDetails().AmenityFurniture = prop.GetDetails().GetAmenityFurniture()
		}

		prop = nprop
	}

	if prop.GetDetails().GetType() == property.PropertyType_PROPERTY_TYPE_STUDIO &&
		prop.GetDetails().BedroomsNo == nil {
		prop.GetDetails().BedroomsNo = wrapperspb.Int32(0)
	}

	return prop, nil
}

func (srv *serviceImpl) rangeOffers(ctx context.Context, prop *extras.Property) *offer.Offer {
	resp, err := srv.config.Connections.Offer.ListByProperty(grpcctx.WithServiceUser(ctx), &offer.ListRequest{
		Property: &offer.ListRequest_PropertyValue{PropertyValue: prop.GetId()},
	})
	if err != nil {
		return nil
	}

	for _, v := range resp.GetOffers() {
		found := false

		for _, t := range prop.GetPeople().GetTenants().GetValue() {
			if t.GetContract().GetValue() == v.GetContract() {
				found = true
				break
			}
		}

		if !found {
			continue
		}

		return v
	}

	return nil
}
